/* 
   color.cpp

   RGB-color (+alpha-channel)
 
     Copyright (C) 2005, Patrick Rammelt <rammelt@cs.tu-berlin.de>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
*/

#ifndef _COLOR_CPP
#define _COLOR_CPP


#include "mathematical.cpp"


#ifndef _NOTIFF
#include "tiffio.h"
#else
typedef unsigned int uint32;
#endif


/* ------------------------------------------------------------------------
   COLOR
   ------------------------------------------------------------------------ */
class color 
{
protected:

  double cred;   // red   [0..255]
  double cgreen; // green [0..255]
  double cblue;  // blue  [0..255]
  double calpha; // alpha [0..255]

public:

  // construct
  color (void) : cred(0), cgreen(0), cblue(0), calpha(0) { }
  color (const class color &c) { set (c); }
  color (double gray, double a=0.0) { set (gray,a); }
  color (double r, double g, double b, double a=0.0) { set (r,g,b,a); }
  color (uint32 c) { set(c); }
  virtual ~color (void) {}

  // set
  virtual inline color& set (double gray, double a=0.0) {
    cred   = trim (gray,0,255); 
    cgreen = trim (gray,0,255); 
    cblue  = trim (gray,0,255); 
    calpha = trim (a,0,255); 
    return (*this);
  }
  virtual inline color& set (double r, double g, double b, double a=0.0) { 
    cred   = trim (r,0,255); 
    cgreen = trim (g,0,255); 
    cblue  = trim (b,0,255); 
    calpha = trim (a,0,255); 
    return (*this);
  }
  virtual inline color& set (uint32 c) { 
    cred   = (double)(getR(c)); 
    cgreen = (double)(getG(c)); 
    cblue  = (double)(getB(c)); 
    calpha = (double)(getA(c)); 
    return (*this);
  }
  virtual inline color& set (const class color &c) {
    cred   = c.cred; 
    cgreen = c.cgreen; 
    cblue  = c.cblue; 
    calpha = c.calpha; 
    return (*this);
  }

  // set colors
  inline void red     (double r) { cred   = trim (r,0,255); }
  inline void green   (double g) { cgreen = trim (g,0,255); }
  inline void blue    (double b) { cblue  = trim (b,0,255); }
  inline void alpha   (double a) { calpha = trim (a,0,255); }
  inline void grey    (double v) { set (v); }
  inline void channel (double v, unsigned int c) { 
    switch (c) {
    case 0:  cred   = trim (v,0,255); break;
    case 1:  cgreen = trim (v,0,255); break;
    case 2:  cblue  = trim (v,0,255); break;
    default: calpha = trim (v,0,255); break; } 
  }

  // get colors
  inline double red   (void) const { return (cred);   }
  inline double green (void) const { return (cgreen); }
  inline double blue  (void) const { return (cblue);  }
  inline double alpha (void) const { return (calpha); }
  inline double grey  (void) const { return ((cred+cgreen+cblue)/(double)3.0);}
  inline double channel (unsigned int c) { 
    switch (c) {
    case 0:  return (cred);
    case 1:  return (cgreen);
    case 2:  return (cblue);
    default: return (calpha); } 
  }

  // get uint32 (tiff) color format
  inline uint32 tiff (void) const { 
    return ((((uint32)(trim (cred,  0,255)) & 0xff))     |
	    (((uint32)(trim (cgreen,0,255)) & 0xff)<<8)  | 
	    (((uint32)(trim (cblue, 0,255)) & 0xff)<<16) | 
	    (((uint32)(trim (calpha,0,255)) & 0xff)<<24)); 
  }

  // enlight red,green and blue all by single factor l 
  class color& lighting (double l) { 
    if (l<0.0) {
      set(0,0,0); 
    } else if (l<=1.0) { 
      cred   *= l; 
      cgreen *= l; 
      cblue  *= l; 
    } 
    return (*this); 
  }

  // enlight red,green and blue by factors r, g and b respectively
  class color& lighting (double r, double g, double b) { 
    if (r<0.0) cred   = 0; else if (r<=1.0) cred   *= r; 
    if (g<0.0) cgreen = 0; else if (g<=1.0) cgreen *= g; 
    if (b<0.0) cblue  = 0; else if (b<=1.0) cblue  *= b; 
    return (*this);
  }

  // reflecting c by reflex-coefficient (0:none 1:full)
  class color& reflecting (const class color &c, double reflex) {
    lighting (((double)c.cred  /255.0*reflex) + (1.0-reflex),
	      ((double)c.cgreen/255.0*reflex) + (1.0-reflex),
	      ((double)c.cblue /255.0*reflex) + (1.0-reflex));
    return (*this);
  }
 
  // set
  inline class color& operator = (const class color &c) { 
    return (set (c));  
  }
  inline class color& operator = (double gray) { 
    return (set (gray)); 
  }

  // color "multiplication" (->lighting)
  inline class color& operator *= (double l) { 
    return (lighting (l)); 
  }
  inline class color& operator *= (const class color &c) { 
    return (lighting ((double)c.cred  /(double)255,
		      (double)c.cgreen/(double)255,
		      (double)c.cblue /(double)255)); 
  }
  inline class color operator * (const class color &c) const {
    class color x (*this); 
    x *= c; 
    return (x); 
  }
  inline class color operator * (double l) const {
    class color x (*this); 
    x *= l; 
    return (x); 
  }

  // if c is outside [a,b] return the nearest value (a or b respectively)
  inline double trim (double c, double a, double b) const {
    if (c<a)      return a;
    else if (c>b) return b;
    else          return c;
  }

  // get separated color-channels from single uint32-value
  inline uint32 getR (uint32 a) const { return ((a) & 0xff); }
  inline uint32 getG (uint32 a) const { return (((a) >> 8) & 0xff); }
  inline uint32 getB (uint32 a) const { return (((a) >> 16) & 0xff); }
  inline uint32 getA (uint32 a) const { return (((a) >> 24) & 0xff); }
};


#endif // ifndef _COLOR_CPP
