/* 
   image.cpp

   image (2d-pixel-field)
 
     Copyright (C) 2005, Patrick Rammelt <rammelt@cs.tu-berlin.de>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
*/

#ifndef _IMAGE_CPP
#define _IMAGE_CPP


#include <stdlib.h>


/* ========================================================================
   IMAGE
   ======================================================================== */
template <class valuetype>
class image
{
protected:

  valuetype *v;  // value field
  long w;        // width
  long h;        // height

  double x0;     // left
  double y0;     // top
  double x1;     // right
  double y1;     // bottom

  valuetype nov; // no value

public:

  // construct / destruct
  image (void) { init(); }
  image (const class image &img) { 
    this->v = NULL; 
    this->w = 0; 
    this->h = 0; 
    set (img); 
  }
  image (long wid, long hei) { init (wid,hei); }
  virtual ~image (void) { del(); }

  // reconstruct / delete
  virtual void del (void) { 
    if (v!=NULL) { delete[] v; } v=NULL; w=0; h=0; 
  }
  virtual void set (long wid, long hei) { 
    if (wid!=w || h!=hei) { 
      del(); 
      w = wid; 
      h = hei; 
      if (w*h>0) v = new valuetype[w*h]; 
    }
  }
  virtual void set (const class image &img) { 
    x0 = img.x0; x1 = img.x1; y0 = img.y0; y1 = img.y1; 
    set(img.w,img.h); 
    for (long i=0, s=img.w*img.h; i<s; v[i]=img.v[i], i++); 
  }
  
  // zoom
  virtual void set_zoom (double lft, double tp, double rgt, double bot) {
    x0 = lft; x1 = rgt; y0 = tp; y1 = bot; }
  virtual double left   (void) const { return (x0); }
  virtual double right  (void) const { return (x1); }
  virtual double top    (void) const { return (y0); }
  virtual double bottom (void) const { return (y1); }
  virtual double xzoom  (void) const { return ((double)w/(x1-x0)); }
  virtual double yzoom  (void) const { return ((double)h/(y1-y0)); }
  virtual double xzoom  (double x) const { return (x/xzoom()+left()); }
  virtual double yzoom  (double y) const { return (y/yzoom()+top());  }

  // set "no-val" (value returned for x-y coordinates outside image)
  virtual void set_noval (const valuetype &val) { nov=val; }

  // set value
  virtual void set_all (const valuetype &val) { 
    for (long i=0; i<w*h; v[i]=val, i++); 
  }
  virtual bool set_xy (long x, long y, const valuetype &val) {
    if (isin(x,y)) { v[x+y*w]=val; return (1); } else return (0);
  }
  virtual bool set_xyconv (double x, double y, const valuetype &val) {
    return (set_xy(xconv(x),yconv(y),val)); 
  }
  
  // get value
  virtual valuetype get_xy (long x, long y) const { 
    if (isin (x,y)) return (v[x+y*w]); else return (nov); 
  }
  virtual valuetype get_xyconv (long x, long y) const { 
    return (get_xy (xconv(x),yconv(y))); 
  }
  virtual int get_xy (long x, long y, valuetype &val) const { 
    if (isin(x,y)) { val = v[x+y*w]; return(1); } 
    else { val = nov; return(0); } 
  }
  virtual int get_xyconv (long x, long y, valuetype &val) const { 
    return (get_xyconv(x,y,val)); 
  }
  
  // get pointer to value
  virtual valuetype* ref_xy (long x, long y) const {
    if (isin(x,y)) return (&v[x+y*w]); else return (NULL); 
  }
  virtual valuetype* ref_xyconv (long x, long y) const {
    return (ref_xy(xconv(x),yconv(y))); 
  }
  
  // x, y inside
  bool isin (long x, long y) const { return (x>=0 && x<w && y>=0 && y<h); }
  
  // get width / height 
  long width  (void) const { return (w); }
  long height (void) const { return (h); }

  // convert x/y-value to pixel-nr
  long xconv (double x) const { return ((long)((x-x0)/(x1-x0)*(double)w)); }
  long yconv (double y) const { return ((long)((y-y0)/(y1-y0)*(double)h)); }

  // getting ugly here...
  virtual valuetype* field (void) const { return (v); }
  
protected:

  virtual void init (void) { v=NULL; w=0; h=0; x0=x1=y0=y1=0; }
  virtual void init (long wid, long hei) { 
    init(); x0=0; y0=0; x1=wid; y1=hei; set(wid,hei); 
  }
};


#endif // ifndef _IMAGE_CPP
