/* 
   photoimage.cpp

   Easy interface to libjpeg and libtiff.
 
     Copyright (C) 2005, Patrick Rammelt <rammelt@cs.tu-berlin.de>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
*/

#ifndef _PHOTOIMAGE_CPP
#define _PHOTOIMAGE_CPP

#include <string.h>  // (memcpy)
#include "image.cpp"
#include "color.cpp"

#ifndef _NOTIFF
#include "tiffio.h"
#endif

#ifndef _NOJPEG
extern "C"
{
#include <jpeglib.h>
}
#endif


#define PHOTOIMAGE_TIFF 1
#define PHOTOIMAGE_JPEG 2


/* ========================================================================
   TIFF IMAGE
   ======================================================================== */
class photoimage : public image<uint32> {

public:

  // construct
  photoimage (void) { init(); }
  photoimage (long wid, long hei) { init(wid,hei); }
  photoimage (long wid, long hei, const class color &c) { 
    init(wid,hei); paint(c); 
  }
  photoimage (char *filename, int type) { v=NULL; read (filename,type); }

  // read from file
  int read (char *filename, int type) {
    switch (type) {
    case PHOTOIMAGE_TIFF: return (read_tiff(filename)); 
    case PHOTOIMAGE_JPEG: return (read_jpeg(filename));
    default:              return (0);
    }
  }
  int read_tiff  (char *filename);
  int read_jpeg  (char *filename);

  // write to file 
  int write (char *filename, int type) {           // with default parameters
    switch (type) {
    case PHOTOIMAGE_TIFF: return (write_tiff(filename)); 
    case PHOTOIMAGE_JPEG: return (write_jpeg(filename));
    default:              return (0);
    }
  }
  int write_tiff (char *filename, bool alpha=1);   // (incl. alpha-channel?)
  int write_jpeg (char *filename, int quality=75); 

  // reconstruct / delete
  virtual void del (void) { 
    if (v!=NULL) { free((uint32*)v); } v=NULL; w=0; h=0; }
  virtual void set (long wid, long hei) { 
    if (wid!=w || h!=hei) { 
      del(); w=wid; h=hei; 
      if (w*h>0) v=(uint32*)malloc(w*h*sizeof(uint32)); 
    } 
  }
  virtual void set (const class photoimage &img) { image<uint32>::set (img); }

  // get / set color of one pixel / all pixels
  class color pixel (long x, long y) const { return (get_xy(x,y)); }
  virtual void paint (long x, long y, 
		      const class color &c) { set_xy(x,y,c.tiff());  }
  virtual void paint (class color c) { uint32 col=c.tiff(); set_all(col); }

  class color operator () (long x, long y) const { return (get_xy(x,y)); }
  void operator () (long x, long y, const class color &c) { 
    set_xy(x,y,c.tiff()); 
  }

};


//        _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ //
// TIFF  _/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/ //
//                                                                          //

#ifndef _NOTIFF 


/* ------------------------------------------------------------------------
   read tiff-image from file
   ------------------------------------------------------------------------ */
int photoimage::read_tiff (char *filename)
{
  TIFF *img;
  TIFFRGBAImage rgbimg[1];
  char errmsg[1024];

  if ((img = (TIFF*) TIFFOpen (filename,"r"))==NULL) {
    printf ("ERROR CAN NOT OPEN FILE %s\n",filename); 
    return (0); }

  uint16 compmode; 
  uint32 rowsperstrip; 
  TIFFGetField (img,TIFFTAG_COMPRESSION,&compmode);
  TIFFGetField (img,TIFFTAG_ROWSPERSTRIP,&rowsperstrip);

  if (!TIFFRGBAImageOK(img,errmsg)) {
    printf ("%s\n",errmsg);
    return (0); }

  if (!TIFFRGBAImageBegin(rgbimg,img,1,errmsg)) {
    printf ("%s\n",errmsg);
    return (0); }

  set (rgbimg->width,rgbimg->height);

  if (!TIFFRGBAImageGet(rgbimg,v,(uint32)w,(uint32)h)) {
    printf ("%s\n",errmsg);
    return (0); }

  return (1);
}


/* ------------------------------------------------------------------------
   write tiff-image to file
   ------------------------------------------------------------------------ */
int photoimage::write_tiff (char *filename, bool alpha) 
{
  TIFF *outimg;

  // open image file
  if ((outimg = (TIFF*) TIFFOpen (filename,"w"))==NULL) return (0);

  const int spp = (3+(int)alpha); // samples per pixel RGB+alpha

  // set tiff-tags
  TIFFSetField (outimg,TIFFTAG_SAMPLEFORMAT,SAMPLEFORMAT_UINT);
  TIFFSetField (outimg,TIFFTAG_ROWSPERSTRIP,
		TIFFDefaultStripSize(outimg, w*spp));
  TIFFSetField (outimg,TIFFTAG_IMAGEWIDTH,w);
  TIFFSetField (outimg,TIFFTAG_IMAGELENGTH,h);
  TIFFSetField (outimg,TIFFTAG_PLANARCONFIG,PLANARCONFIG_CONTIG);
  TIFFSetField (outimg,TIFFTAG_BITSPERSAMPLE,8);
  TIFFSetField (outimg,TIFFTAG_SAMPLESPERPIXEL,spp); 
  TIFFSetField (outimg,TIFFTAG_PHOTOMETRIC,PHOTOMETRIC_RGB);
  TIFFSetField (outimg,TIFFTAG_COMPRESSION,COMPRESSION_NONE);
  if (alpha) {
    uint16 extra[1] = {EXTRASAMPLE_UNASSALPHA}; 
    TIFFSetField(outimg, TIFFTAG_EXTRASAMPLES, 1, extra);
  }
  // TIFFSetField (outimg,TIFFTAG_COMPRESSION,COMPRESSION_JPEG);
  // TIFFSetField (outimg,TIFFTAG_JPEGPROC,JPEGPROC_LOSSLESS);
  // TIFFSetField (outimg,TIFFTAG_ORIENTATION,ORIENTATION_LEFTBOT);

  // write image file
  const unsigned int bpp = spp*sizeof(unsigned char);     // bytes per pixel
  const unsigned int bpl = w*bpp;                         // bytes per line
  unsigned char *buf = (unsigned char*)malloc(bpl);       // line-buffer
  if (alpha) {                                    // including alpha-channel
    for (long i=0; i<(long)h; i++) {
      memcpy (buf,&v[i*w],bpl);                   // correct endian-order
      TIFFWriteScanline (outimg,buf,i,(tsample_t)0); 
    }
  } else {                                        // excluding alpha-channel
    for (long i=0; i<(long)h; i++) {
      for (long j=0; j<w; j++) {
	buf[j*spp+0] = TIFFGetR(v[i*w+j]);        // red
	buf[j*spp+1] = TIFFGetG(v[i*w+j]);        // green
	buf[j*spp+2] = TIFFGetB(v[i*w+j]);        // blue
      }
      TIFFWriteScanline (outimg,buf,i,(tsample_t)0); 
    }
  }
  free(buf);

  // close image file
  TIFFClose ((TIFF*)outimg);
  return (1);
}


#else  // ifndef _NOTIFF

/* read / write dummies  */
int photoimage::read_tiff (char *filename) { return (0); }
int photoimage::write_tiff (char *filename, bool alpha) { return (0); }

#endif // ifndef _NOTIFF


//        _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ //
// JPEG  _/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/ //
//                                                                          //

#ifndef _NOJPEG 

/* ------------------------------------------------------------------------
   read jpeg-image
   ------------------------------------------------------------------------ */
int photoimage::read_jpeg  (char *filename) 
{ 
  // main struct / error-handler
  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;
  cinfo.err = jpeg_std_error(&jerr);       // set default error-handler
  jpeg_create_decompress(&cinfo);          // ( after error-handler set-up )

  // open input
  FILE * infile;
  if ((infile = fopen(filename, "rb")) == NULL) return (0);

  // read header
  jpeg_stdio_src (&cinfo, infile);
  jpeg_read_header (&cinfo, TRUE);

  // set attributes for decompression (after jpeg_read_header)
  cinfo.out_color_space = JCS_RGB;                       // rgb-image wanted

  // create object
  set (cinfo.image_width,cinfo.image_height);

  // decompress
  jpeg_start_decompress (&cinfo);
  unsigned int i;
  const unsigned int spp = cinfo.output_components;      // samples per pixel
  JSAMPARRAY buf = ((*cinfo.mem->alloc_sarray)
		    ((j_common_ptr) &cinfo, JPOOL_IMAGE, w*spp, 1));
  while ((i=cinfo.output_scanline) < cinfo.output_height) {
    jpeg_read_scanlines(&cinfo, buf, 1);
    for (long j=0; j<w; j++) {
      v[i*w+j] = ((((uint32)buf[0][j*spp+0] & 0xff)) |      // red
		  (((uint32)buf[0][j*spp+1] & 0xff)<<8) |   // green
		  (((uint32)buf[0][j*spp+2] & 0xff)<<16));  // blue
    }
  }
  jpeg_finish_decompress (&cinfo);

  // close input / free memory
  fclose (infile);
  jpeg_destroy_decompress (&cinfo);
  return (1);
}


/* ------------------------------------------------------------------------
   write jpeg-image to file
   ------------------------------------------------------------------------ */
int photoimage::write_jpeg (char *filename, int quality) 
{
  // main struct / error-handler
  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;
  cinfo.err = jpeg_std_error (&jerr);      // set default error-handler
  jpeg_create_compress (&cinfo);           // ( after error-handler set-up )

  // open output
  FILE * outfile;
  if ((outfile = fopen(filename, "wb")) == NULL) return (0);
  jpeg_stdio_dest (&cinfo, outfile);

  // set attributes
  cinfo.image_width      = w;
  cinfo.image_height     = h;
  cinfo.input_components = 3;
  cinfo.in_color_space   = JCS_RGB;
  jpeg_set_defaults (&cinfo);     // ( after set-up of cinfo.in_color_space )

  // compress
  unsigned int i;
  const unsigned int spp = 3;                         // samples per pixel RGB
  const unsigned int bpp = spp*sizeof(unsigned char); // bytes per pixel
  const unsigned int bpl = w*bpp;                     // bytes per line
  unsigned char *buf = (unsigned char*)malloc(bpl);   // line-buffer
  jpeg_start_compress (&cinfo, TRUE);
  while ((i=cinfo.next_scanline) < cinfo.image_height) {
    for (long j=0; j<w; j++) {
      buf[j*spp+0] = (v[i*w+j])       & 0xff;         // red
      buf[j*spp+1] = (v[i*w+j] >>  8) & 0xff;         // green
      buf[j*spp+2] = (v[i*w+j] >> 16) & 0xff;         // blue
    }
    jpeg_write_scanlines (&cinfo, &buf, 1);
  }
  jpeg_finish_compress (&cinfo);
  free (buf);

  // close output / free memory
  fclose (outfile);
  jpeg_destroy_compress (&cinfo);
  return (1);
}


#else // ifndef _NOJPEG

/* dummies */
int photoimage::read_jpeg  (char *filename) { return (0); }
int photoimage::write_jpeg (char *filename, int quality) { return (0); }

#endif // ifndef _NOJPEG


#endif // ifndef _PHOTOIMAGE_CPP
